import os
from unittest import mock

import pytest
import stringcase

from openhexa.sdk.workspaces.workspace import ConnectionDoesNotExist, workspace


def test_workspace_dhis2_connection_not_exist():
    slug = "polio-ff3a0d"
    with pytest.raises(ConnectionDoesNotExist):
        workspace.dhis2_connection(slug=slug)


def test_workspace_dhis2_connection():
    slug = "polio-ff3a0d"
    env_variable_prefix = stringcase.constcase(slug)
    url = "https://test.dhis2.org/"
    username = "dhis2"
    password = "dhis2_pwd"
    with mock.patch.dict(
        os.environ,
        {
            f"{env_variable_prefix}_URL": url,
            f"{env_variable_prefix}_USERNAME": username,
            f"{env_variable_prefix}_PASSWORD": password,
        },
    ):
        dhis2_connection = workspace.dhis2_connection(slug=slug)
        assert dhis2_connection.url == url
        assert dhis2_connection.username == username
        assert dhis2_connection.password == password


def test_workspace_postgresql_connection_not_exist():
    slug = "polio-ff3a0d"
    with pytest.raises(ConnectionDoesNotExist):
        workspace.postgresql_connection(slug=slug)


def test_workspace_postgresql_connection():
    slug = "polio-ff3a0d"
    env_variable_prefix = stringcase.constcase(slug)
    host = "https://172.17.0.1"
    port = "5432"
    username = "dhis2"
    password = "dhis2_pwd"
    database_name = "polio"
    url = f"postgresql://{username}:{password}@{host}:{port}/{database_name}"
    with mock.patch.dict(
        os.environ,
        {
            f"{env_variable_prefix}_HOST": host,
            f"{env_variable_prefix}_USERNAME": username,
            f"{env_variable_prefix}_PASSWORD": password,
            f"{env_variable_prefix}_PORT": port,
            f"{env_variable_prefix}_DB_NAME": database_name,
        },
    ):
        postgres_connection = workspace.postgresql_connection(slug=slug)
        assert postgres_connection.username == username
        assert postgres_connection.password == password
        assert postgres_connection.host == host
        assert postgres_connection.port == int(port)
        assert postgres_connection.database_name == database_name
        assert postgres_connection.url == url


def test_workspace_S3_connection_not_exist():
    slug = "polio-ff3a0d"
    with pytest.raises(ConnectionDoesNotExist):
        workspace.s3_connection(slug=slug)


def test_workspace_s3_connection():
    slug = "polio-ff3a0d"
    env_variable_prefix = stringcase.constcase(slug)
    secret_access_key = "HqQBxH0BAI3zF7kANUNlGg"
    access_key_id = "84hVntMaMSYP/RSW9ex04w"
    bucket_name = "test"

    with mock.patch.dict(
        os.environ,
        {
            f"{env_variable_prefix}_SECRET_ACCESS_KEY": secret_access_key,
            f"{env_variable_prefix}_ACCESS_KEY_ID": access_key_id,
            f"{env_variable_prefix}_BUCKET_NAME": bucket_name,
        },
    ):
        s3_connection = workspace.s3_connection(slug=slug)
        assert s3_connection.secret_access_key == secret_access_key
        assert s3_connection.access_key_id == access_key_id
        assert s3_connection.bucket_name == bucket_name


def test_workspace_gcs_connection_not_exist():
    slug = "polio-ff3a0d"
    with pytest.raises(ConnectionDoesNotExist):
        workspace.gcs_connection(slug=slug)


def test_workspace_gcs_connection():
    slug = "polio-ff3a0d"
    env_variable_prefix = stringcase.constcase(slug)
    service_account_key = "HqQBxH0BAI3zF7kANUNlGg"
    bucket_name = "test"

    with mock.patch.dict(
        os.environ,
        {
            f"{env_variable_prefix}_SERVICE_ACCOUNT_KEY": service_account_key,
            f"{env_variable_prefix}_BUCKET_NAME": bucket_name,
        },
    ):
        s3_connection = workspace.gcs_connection(slug=slug)
        assert s3_connection.service_account_key == service_account_key
        assert s3_connection.bucket_name == bucket_name


def test_workspace_custom_connection():
    slug = "polio-ff3a0d"
    env_variable_prefix = stringcase.constcase(slug)
    username = "kaggle_username"
    password = "root"

    with mock.patch.dict(
        os.environ,
        {
            f"{env_variable_prefix}_USERNAME": username,
            f"{env_variable_prefix}_PASSWORD": password,
        },
    ):
        custom_connection = workspace.custom_connection(slug=slug)
        assert len(custom_connection.fields) == 2
        assert f"{env_variable_prefix}_USERNAME" in custom_connection.fields
        assert f"{env_variable_prefix}_PASSWORD" in custom_connection.fields



import base64
import configparser
import io
import os
from pathlib import Path
from zipfile import ZipFile

import click
import requests

from openhexa import __version__
from openhexa.sdk.pipelines import import_pipeline

CONFIGFILE_PATH = os.path.expanduser("~") + "/.openhexa.ini"


def is_debug(config: configparser.ConfigParser):
    return config.getboolean("openhexa", "debug", fallback=False)


def open_config():
    config = configparser.ConfigParser()
    if os.path.exists(CONFIGFILE_PATH):
        config.read(CONFIGFILE_PATH)
    else:
        config.read_string(
            """
        [openhexa]
        url=https://api.openhexa.org

        [workspaces]
        """
        )
    return config


def save_config(config):
    with open(CONFIGFILE_PATH, "w") as configfile:
        config.write(configfile)


def graphql(config, query: str, variables=None, token=None):
    url = config["openhexa"]["url"] + "/graphql/"
    current_workspace = config["openhexa"]["current_workspace"]
    if token is None:
        token = config["workspaces"].get(current_workspace)

    if not token:
        raise Exception("No token found for workspace")

    if is_debug(config):
        click.echo("")
        click.echo("Graphql Query:")
        click.echo(f"URL: {url}")
        click.echo(f"Query: {query}")
        click.echo(f"Variables: {variables}")

    response = requests.post(
        url,
        headers={
            "User-Agent": f"openhexa-cli/{__version__}",
            "Authorization": f"Bearer {token}",
        },
        json={"query": query, "variables": variables},
    )
    response.raise_for_status()
    data = response.json()

    if is_debug(config):
        click.echo("Graphql Response:")
        click.echo(data)
        click.echo("")

    if data.get("errors"):
        if data.get("errors")[0].get("extensions", {}).get("code") == "UNAUTHENTICATED":
            raise Exception("Token is invalid, please update the token")
        raise Exception(data["errors"])
    return data["data"]


def get_workspace(config, slug: str, token: str):
    return graphql(
        config,
        """
            query getWorkspace($slug: String!) {
                workspace(slug: $slug) {
                    name
                    slug
                }
            }
            """,
        {"slug": slug},
        token,
    )["workspace"]


def get_pipelines(config):
    data = graphql(
        config,
        """
    query getWorkspacePipelines($workspaceSlug: String!) {
        pipelines(workspaceSlug: $workspaceSlug) {
            items {
                id
                code
                name
                currentVersion {
                    number
                }
            }
        }
    }
    """,
        {"workspaceSlug": config["openhexa"]["current_workspace"]},
    )
    return data["pipelines"]["items"]


def get_pipeline(config, pipeline_code: str):
    data = graphql(
        config,
        """
    query getWorkspacePipeline($workspaceSlug: String!, $pipelineCode: String!) {
        pipelineByCode (workspaceSlug: $workspaceSlug, code: $pipelineCode) {
            id
            code
            currentVersion {
                number
            }
        }
    }
    """,
        {
            "workspaceSlug": config["openhexa"]["current_workspace"],
            "pipelineCode": pipeline_code,
        },
    )
    return data["pipelineByCode"]


def create_pipeline(config, pipeline_code: str, pipeline_name: str):
    data = graphql(
        config,
        """
    mutation createPipeline($input: CreatePipelineInput!) {
        createPipeline(input: $input) {
            success
            errors
            pipeline {
                id
                code
                name
            }
        }
    }
    """,
        {
            "input": {
                "workspaceSlug": config["openhexa"]["current_workspace"],
                "code": pipeline_code,
                "name": pipeline_name,
            }
        },
    )

    if not data["createPipeline"]["success"]:
        raise Exception(data["createPipeline"]["errors"])

    return data["createPipeline"]["pipeline"]


def ensure_is_pipeline_dir(pipeline_path: str):
    # Ensure that there is a pipeline.py file in the directory
    if not os.path.isdir(pipeline_path):
        raise Exception(f"Path {pipeline_path} is not a directory")
    if not os.path.exists(pipeline_path):
        raise Exception(f"Directory {pipeline_path} does not exist")
    if not os.path.exists(os.path.join(pipeline_path, "pipeline.py")):
        raise Exception(
            f"Directory {pipeline_path} does not contain a pipeline.py file"
        )

    return True


def upload_pipeline(config, pipeline_directory_path: str):
    pipeline = import_pipeline(pipeline_directory_path)
    directory = Path(os.path.abspath(pipeline_directory_path))

    zipFile = io.BytesIO(b"")

    if is_debug(config):
        click.echo("Generating ZIP file:")

    with ZipFile(zipFile, "w") as zipObj:
        for path in directory.glob("**/*"):
            if not path.is_file():
                continue
            if not path.name.endswith(".py") and not path.name == "requirements.txt":
                continue
            if is_debug(config):
                click.echo(f"\t{path.name}")
            zipObj.write(path, path.relative_to(directory))

    zipFile.seek(0)

    if is_debug(config):
        # Write zipFile to disk for debugging
        open("pipeline.zip", "wb").write(zipFile.read())
        zipFile.seek(0)

    base64_content = base64.b64encode(zipFile.read()).decode("ascii")

    data = graphql(
        config,
        """
            mutation uploadPipeline($input: UploadPipelineInput!) {
                uploadPipeline(input: $input) {
                    success
                    errors
                    version
                }
            }
        """,
        {
            "input": {
                "workspaceSlug": config["openhexa"]["current_workspace"],
                "code": pipeline.code,
                "zipfile": base64_content,
                "parameters": pipeline.parameters_spec(),
            }
        },
    )

    if not data["uploadPipeline"]["success"]:
        raise Exception(data["uploadPipeline"]["errors"])

    return data["uploadPipeline"]["version"]



import dataclasses


@dataclasses.dataclass
class DHIS2Connection:
    url: str
    username: str
    password: str


@dataclasses.dataclass
class PostgreSQLConnection:
    host: str
    port: int
    username: str
    password: str
    database_name: str

    @property
    def url(self):
        return (
            f"postgresql://{self.username}:{self.password}"
            f"@{self.host}:{self.port}/{self.database_name}"
        )


@dataclasses.dataclass
class S3Connection:
    access_key_id: str
    secret_access_key: str
    bucket_name: str


@dataclasses.dataclass
class GCSConnection:
    service_account_key: str
    bucket_name: str


@dataclasses.dataclass
class CustomConnection:
    fields: dict



import os
import re

import stringcase

from openhexa.sdk.utils import Environments, get_environment

from .connection import (
    CustomConnection,
    DHIS2Connection,
    GCSConnection,
    PostgreSQLConnection,
    S3Connection,
)


class WorkspaceConfigError(Exception):
    pass


class ConnectionDoesNotExist(Exception):
    pass


class CurrentWorkspace:
    @property
    def database_host(self):
        try:
            return os.environ["WORKSPACE_DATABASE_HOST"]
        except KeyError:
            raise WorkspaceConfigError(
                "No database has been configured. Did you forget to provide a database entry"
                " in your workspace.yaml file?"
            )

    @property
    def database_username(self):
        try:
            return os.environ["WORKSPACE_DATABASE_USERNAME"]
        except KeyError:
            raise WorkspaceConfigError(
                "No database has been configured. Did you forget to provide a database entry"
                " in your workspace.yaml file?"
            )

    @property
    def database_password(self):
        try:
            return os.environ.get("WORKSPACE_DATABASE_PASSWORD")
        except KeyError:
            raise WorkspaceConfigError(
                "No database has been configured. Did you forget to provide a database entry"
                " in your workspace.yaml file?"
            )

    @property
    def database_name(self):
        try:
            return os.environ["WORKSPACE_DATABASE_DB_NAME"]
        except KeyError:
            raise WorkspaceConfigError(
                "No database has been configured. Did you forget to provide a database entry"
                " in your workspace.yaml file?"
            )

    @property
    def database_port(self):
        try:
            return int(os.environ["WORKSPACE_DATABASE_PORT"])
        except KeyError:
            raise WorkspaceConfigError(
                "No database has been configured. Did you forget to provide a database entry"
                " in your workspace.yaml file?"
            )

    @property
    def database_url(self):
        return (
            f"postgresql://{self.database_username}:{self.database_password}"
            f"@{self.database_host}:{self.database_port}/{self.database_name}"
        )

    @property
    def files_path(self) -> str:
        """Return the base path to the filesystem, without trailing slash"""
        env = get_environment()
        if env == Environments.LOCAL:
            if "WORKSPACE_FILES_PATH" not in os.environ:
                raise WorkspaceConfigError(
                    "No filesystem has been configured. Did you forget to provide a files.path"
                    "key in your workspace.yaml file?"
                )
            return os.environ["WORKSPACE_FILES_PATH"]
        else:
            return "/home/hexa/workspace"

    def dhis2_connection(self, slug: str) -> DHIS2Connection:
        try:
            env_variable_prefix = stringcase.constcase(slug.lower())
            url = os.environ[f"{env_variable_prefix}_URL"]
            username = os.environ[f"{env_variable_prefix}_USERNAME"]
            password = os.environ[f"{env_variable_prefix}_PASSWORD"]
        except KeyError:
            raise ConnectionDoesNotExist(f'No DHIS2 connection for "{slug}"')

        return DHIS2Connection(url=url, username=username, password=password)

    def postgresql_connection(self, slug: str) -> PostgreSQLConnection:
        try:
            env_variable_prefix = stringcase.constcase(slug.lower())
            host = os.environ[f"{env_variable_prefix}_HOST"]
            port = int(os.environ[f"{env_variable_prefix}_PORT"])
            username = os.environ[f"{env_variable_prefix}_USERNAME"]
            password = os.environ[f"{env_variable_prefix}_PASSWORD"]
            dbname = os.environ[f"{env_variable_prefix}_DB_NAME"]
        except KeyError:
            raise ConnectionDoesNotExist(f'No PostgreSQL connection for "{slug}"')

        return PostgreSQLConnection(
            host=host,
            port=port,
            username=username,
            password=password,
            database_name=dbname,
        )

    def s3_connection(self, slug: str) -> S3Connection:
        try:
            env_variable_prefix = stringcase.constcase(slug.lower())
            secret_access_key = os.environ[f"{env_variable_prefix}_SECRET_ACCESS_KEY"]
            access_key_id = os.environ[f"{env_variable_prefix}_ACCESS_KEY_ID"]
            bucket_name = os.environ[f"{env_variable_prefix}_BUCKET_NAME"]
        except KeyError:
            raise ConnectionDoesNotExist(f'No S3 connection for "{slug}"')

        return S3Connection(
            access_key_id=access_key_id,
            secret_access_key=secret_access_key,
            bucket_name=bucket_name,
        )

    def gcs_connection(self, slug: str) -> GCSConnection:
        try:
            env_variable_prefix = stringcase.constcase(slug.lower())
            service_account_key = os.environ[
                f"{env_variable_prefix}_SERVICE_ACCOUNT_KEY"
            ]
            bucket_name = os.environ[f"{env_variable_prefix}_BUCKET_NAME"]
        except KeyError:
            raise ConnectionDoesNotExist(f'No GCS connection for "{slug}"')

        return GCSConnection(
            service_account_key=service_account_key,
            bucket_name=bucket_name,
        )

    def custom_connection(slef, slug: str) -> CustomConnection:
        env_variable_prefix = stringcase.constcase(slug.lower())
        fields = {}
        for key, value in os.environ.items():
            if re.match(rf"^{env_variable_prefix}_", key):
                fields[key] = os.environ[key]
        return CustomConnection(fields=fields)


workspace = CurrentWorkspace()



import base64
import importlib
import io
import os
import sys
from zipfile import ZipFile

import requests

from .pipeline import Pipeline


def import_pipeline(pipeline_dir_path: str):
    pipeline_dir = os.path.abspath(pipeline_dir_path)
    sys.path.append(pipeline_dir)
    pipeline_package = importlib.import_module("pipeline")

    pipeline = next(
        v for _, v in pipeline_package.__dict__.items() if v and type(v) == Pipeline
    )
    return pipeline


def download_pipeline(url: str, token: str, run_id: str, target_dir):
    r = requests.post(
        url + "/graphql/",
        headers={"Authorization": f"Bearer {token}"},
        json={
            "query": """
            query PipelineDownload($id: UUID!) {
              pipelineRun(id: $id) {
                id
                version {
                  number
                }
                code
              }
            }
            """,
            "variables": {"id": run_id},
        },
        timeout=30,
    )
    r.raise_for_status()
    data = r.json()
    zipfile = base64.b64decode(data["data"]["pipelineRun"]["code"].encode("ascii"))
    source_dir = os.getcwd()
    os.chdir(target_dir)
    with ZipFile(io.BytesIO(zipfile)) as zf:
        zf.extractall()
    os.chdir(source_dir)



from __future__ import annotations

import datetime
import typing

import openhexa.sdk.pipelines.pipeline


class TaskCom:
    def __init__(self, task):
        self.result = task.result
        self.start_time = task.start_time
        self.end_time = task.end_time


class Task:
    def __init__(self, function: typing.Callable):
        self.name = function.__name__
        self.compute = function
        self.inputs = []
        self.result = None
        self.start_time = None
        self.end_time = None
        self.task_args = {}
        self.task_kwargs = {}
        self.active = False
        self.pooled = False

    def __call__(self, *task_args, **task_kwargs):
        self.active = True  # uncalled tasks will be skipped
        # check that all inputs are tasks
        self.task_args = task_args
        self.task_kwargs = task_kwargs
        return self

    def __repr__(self):
        return self.name

    def get_node_inputs(self):
        inputs = []
        for a in self.task_args:
            if issubclass(type(a), Task):
                inputs.append(a)
        for k, a in self.task_kwargs.items():
            if issubclass(type(a), Task):
                inputs.append(a)
        return inputs

    def is_ready(self):
        if not self.active:
            return False

        for a in self.task_args:
            if issubclass(type(a), Task) and a.end_time is None:
                return False
        for k, a in self.task_kwargs.items():
            if issubclass(type(a), Task) and a.end_time is None:
                return False

        return True if self.end_time is None else False

    def get_tasks_ready(self):
        tasks = []
        for a in self.task_args:
            if issubclass(type(a), Task):
                if a.is_ready():
                    tasks.append(a)
                else:
                    tasks += a.get_tasks_ready()
        for k, a in self.task_kwargs.items():
            if issubclass(type(a), Task):
                if a.is_ready():
                    tasks.append(a)
                else:
                    tasks += a.get_tasks_ready()

        return list(set(tasks))

    def run(self):
        if self.end_time:
            # already executed, return previous result
            return self.result

        # forge task inputs
        r_task_args = []
        for a in self.task_args:
            if issubclass(type(a), Task):
                # previous task; execute + fw results
                r_task_args.append(a.result)
            else:
                # normal parameters, follow up
                r_task_args.append(a)

        r_task_kwargs = {}
        for k, a in self.task_kwargs.items():
            if issubclass(type(a), Task):
                r_task_kwargs[k] = a.result
            else:
                r_task_kwargs[k] = a

        # execute the task
        self.start_time = datetime.datetime.now(datetime.timezone.utc)
        self.result = self.compute(*r_task_args, **r_task_kwargs)
        self.end_time = datetime.datetime.now(datetime.timezone.utc)

        # done!
        return TaskCom(self)

    def stateless_run(self):
        self.result = None
        self.start_time, self.end_time = None, None
        return self.run()


class PipelineWithTask:
    def __init__(
        self,
        function: typing.Callable,
        pipeline: openhexa.sdk.pipelines.pipeline.Pipeline,
    ):
        self.function = function
        self.pipeline = pipeline

    def __call__(self, *task_args, **task_kwargs):
        task = Task(self.function)(*task_args, **task_kwargs)
        self.pipeline.tasks.append(task)
        return task
